#define PLUGIN_VERSION "1.3"

#pragma semicolon 1
#pragma newdecls required

#include <sourcemod>

#define CVAR_FLAGS 			FCVAR_NOTIFY

public Plugin myinfo = 
{
	name = "[ANY] Map Slots Locker",
	author = "Alex Dragokas (idea by rin-ale)",
	description = "Locks ability to join the server on specific maps or by command",
	version = PLUGIN_VERSION,
	url = "https://github.com/dragokas/"
};

const int MAX_MAP_NAME = 64;

char buggedMapsPath[PLATFORM_MAX_PATH];
char g_sCurMap[MAX_MAP_NAME];

int g_iSlotsNumberOld;

bool g_bEnabled;
bool g_bSlotsLocked;
bool g_bSlotsLockedForce;
bool g_bBuggedMap;
bool g_bSpawnHooked;

ArrayList g_hArrayBuggedMaps;

ConVar g_hCvarEnable;
ConVar g_hConVarMaxPlayers;

public void OnPluginStart()
{
	LoadTranslations("maplocker.phrases");
	
	CreateConVar("maplocker_version", PLUGIN_VERSION, "Plugin Version", FCVAR_DONTRECORD | CVAR_FLAGS);
	( g_hCvarEnable = CreateConVar(	"sm_maplocker_enable",		"1",	"Enable the plugin? (1 - Yes, 0 - No)", CVAR_FLAGS )).AddChangeHook(ConVarChanged);
	
	AutoExecConfig(true, "sm_maplocker");
	
	g_hConVarMaxPlayers = FindConVar("sv_maxplayers");
	g_hConVarMaxPlayers.AddChangeHook(ConVarChangedMaxPlayers);
	
	BuildPath(Path_SM, buggedMapsPath, PLATFORM_MAX_PATH, "data/maplocker_list.txt");
	
	g_hArrayBuggedMaps = new ArrayList(ByteCountToCells(MAX_MAP_NAME));
	
	GetCvars();
	
	RegAdminCmd("sm_maplock", 	CmdMapLock, 	ADMFLAG_ROOT, "Lock the map");
	RegAdminCmd("sm_mapunlock", CmdMapUnlock, 	ADMFLAG_ROOT, "Unlock the map");
}

public Action CmdMapLock(int client, int args)
{
	LockSlots();
	g_bSlotsLockedForce = true;
	CPrintToChat(client, "%t", "slots_lock");
	return Plugin_Handled;
}

public Action CmdMapUnlock(int client, int args)
{
	UnlockSlots();
	g_bSlotsLockedForce = false;
	CPrintToChat(client, "%t", "slots_unlock");
	return Plugin_Handled;
}

public void ConVarChanged(ConVar convar, const char[] oldValue, const char[] newValue)
{
	GetCvars();
}

void GetCvars()
{
	g_bEnabled = g_hCvarEnable.BoolValue;

	InitHook();
}

void InitHook()
{
	static bool bHooked;
	
	if( g_bEnabled )
	{
		if( !bHooked ) {
			HookEvent("round_start",		Event_RoundStart,	EventHookMode_PostNoCopy);
			
			bHooked = true;
		}
	} else {
		if( bHooked )
		{
			UnhookEvent("round_start",		Event_RoundStart,	EventHookMode_PostNoCopy);
			bHooked = false;
		}
	}
}

public void OnMapStart()
{
	if( !g_bEnabled )
	{
		return;
	}
	if( g_bSlotsLockedForce )
	{
		return;
	}
	
	static int ft_stri;
	
	int ft = GetFileTime(buggedMapsPath, FileTime_LastChange);
	if( ft != ft_stri )
	{
		ft_stri = ft;
		ReadFileToArrayList(buggedMapsPath, g_hArrayBuggedMaps);
	}
	
	GetCurrentMap(g_sCurMap, sizeof(g_sCurMap));
	
	g_bBuggedMap = IsBuggedMap(g_sCurMap);
}

bool IsBuggedMap(char[] map)
{
	static char badmap[MAX_MAP_NAME];

	for( int i = 0; i < g_hArrayBuggedMaps.Length; i++ )
	{
		g_hArrayBuggedMaps.GetString(i, badmap, sizeof(badmap));
		if( strcmp(badmap, map, false) == 0 )
		{
			return true;
		}
	}
	return false;
}

public void OnMapEnd()
{
	if( !g_bSlotsLockedForce && g_bSlotsLocked )
	{
		UnlockSlots();
	}
}

public void Event_RoundStart(Event event, const char[] name, bool dontBroadcast)
{
	if( !g_bSpawnHooked )
	{
		g_bSpawnHooked = true;
		HookEvent("player_spawn",		Event_PlayerSpawn,	EventHookMode_PostNoCopy);
	}
}

public void Event_PlayerSpawn(Event event, const char[] name, bool dontBroadcast)
{
	int client = GetClientOfUserId(event.GetInt("userid"));
	if( client && IsClientInGame(client) && !IsFakeClient(client) )
	{
		UnhookEvent("player_spawn",	Event_PlayerSpawn,	EventHookMode_PostNoCopy);
		g_bSpawnHooked = false;
		OnFirstPlayerSpawned();
	}
}

void OnFirstPlayerSpawned()
{
	if( g_bBuggedMap && !g_bSlotsLocked )
	{
		CreateTimer(10.0, Timer_LockSlots, _, TIMER_FLAG_NO_MAPCHANGE);
	}
}

public Action Timer_LockSlots(Handle timer)
{
	LockSlots();
	CPrintToChatAll("%t", "slots_lock_mapend");
}

public void ConVarChangedMaxPlayers (ConVar convar, const char[] oldValue, const char[] newValue)
{
	static float fLastTime;
	static int series;
	const int SERIES_MAX = 4;
	
	if( g_bSlotsLocked && strcmp(newValue, "1") != 0 )
	{
		if( GetEngineTime() - fLastTime < 0.1 ) // prevents potential recurse crash due to external hook
		{
			++ series; // compatibility with plugins like "ConVar Anomaly Fixer"
		}
		else {
			series = 0;
		}
		if( series < SERIES_MAX )
		{
			fLastTime = GetEngineTime();
			convar.SetInt(1);
		}
	}
}

void LockSlots()
{
	if( g_iSlotsNumberOld == 0 )
	{
		g_iSlotsNumberOld = g_hConVarMaxPlayers.IntValue;
	}
	g_bSlotsLocked = true;
	g_hConVarMaxPlayers.SetInt(1);
}

void UnlockSlots()
{
	g_bSlotsLocked = false;
	if( g_iSlotsNumberOld != 0 )
	{
		g_hConVarMaxPlayers.SetInt(g_iSlotsNumberOld);
	}
	g_iSlotsNumberOld = 0;
}

stock void ReadFileToArrayList(char[] sPath, ArrayList list)
{
	static char str[MAX_NAME_LENGTH];
	File hFile = OpenFile(sPath, "r");
	if( hFile == null )
	{
		SetFailState("Failed to open file: \"%s\". You are missing at installing!", sPath);
	}
	else {
		list.Clear();
		while( !hFile.EndOfFile() && hFile.ReadLine(str, sizeof(str)) )
		{
			TrimString(str);
			list.PushString(str);
		}
		delete hFile;
	}
}

stock void ReplaceColor(char[] message, int maxLen)
{
	ReplaceString(message, maxLen, "{white}", "\x01", false);
	ReplaceString(message, maxLen, "{cyan}", "\x03", false);
	ReplaceString(message, maxLen, "{orange}", "\x04", false);
	ReplaceString(message, maxLen, "{green}", "\x05", false);
}

stock void CPrintToChat(int iClient, const char[] format, any ...)
{
	static char buffer[192];
	SetGlobalTransTarget(iClient);
	VFormat(buffer, sizeof(buffer), format, 3);
	ReplaceColor(buffer, sizeof(buffer));
	PrintToChat(iClient, "\x01%s", buffer);
}

stock void CPrintToChatAll(const char[] format, any ...)
{
	static char buffer[192];
	for( int i = 1; i <= MaxClients; i++ )
	{
		if( IsClientInGame(i) && !IsFakeClient(i) )
		{
			SetGlobalTransTarget(i);
			VFormat(buffer, sizeof(buffer), format, 2);
			ReplaceColor(buffer, sizeof(buffer));
			PrintToChat(i, "\x01%s", buffer);
		}
	}
}